#$Id: RNA2D.pm,v 1.2 2010/06/17 16:47:03 rnl14480 Exp $
package BAOBAB::RNA2D;

use strict;

sub new {
	my $c = shift;
	my %a = @_;
	my $r = {
		structure => $a{structure},
		sequence  => $a{sequence}
	};
	return bless $r, $c;
}

sub structure { return (shift)->{structure}; }
sub sequence  { return (shift)->{sequence}; }

sub structure_tree {
	my $s = shift;

	unless ( defined $s->{super_compact_tree} ) {
		my $t = $s->_structure2tree();
		my ( $ct, undef ) = $s->_compact_tree( $t, 0 );
		my $sct = $s->_supercompact_tree($ct);

		$s->{super_compact_tree} = $sct;
	}

	return $s->{super_compact_tree};
}

sub _structure2tree {
	my $s = shift;
	my $st = shift || $s->structure();

	my @tree = ();
	my @st = split( //, $st );

	my $level = 0;
	my $sst   = '';

	foreach my $n (@st) {
		if ( $n eq '.' ) {    # loop
			push @tree, $n unless $level;
			$sst .= $n if $level;
		}
		elsif ( $n eq '(' ) {    # left pairing
			$sst .= $n if $level++;
		}
		elsif ( $n eq ')' ) {    #right pairing
			if ( ! --$level ) {
				push @tree, $s->_structure2tree($sst);
				$sst = '';
			}
			else {
				$sst .= $n;
			}
		}
	}

	return \@tree;
}

sub _compact_tree {
	my $s     = shift;
	my $t     = shift;
	my $start = shift || 0;

	my @ct   = ();
	my $len  = 0;
	my $tlen = 0;

	foreach my $n (@$t) {
		if ( $n eq '.' ) {
			$len++;
			$tlen++;
		}
		else {
			push @ct,
			  {
				node  => 'loop',
				start => $start + $tlen - $len,
				stop  => $start + $tlen - 1,
				len   => $len
			  }
			  if $len;

			my ( $sct, $stlen ) = $s->_compact_tree( $n, $start + $tlen + 1 );

			push @ct,
			  {
				node     => 'stem',
				start    => $start + $tlen,
				stop     => $start + $tlen + $stlen + 1,
				len      => $stlen + 2,
				children => $sct
			  };

			$tlen += $stlen + 2;
			$len = 0;
		}
	}

	if ($len) {
		push @ct,
		  {
			node  => 'loop',
			start => $start + $tlen - $len,
			stop  => $start + $tlen - 1,
			len   => $len
		  };
	}

	return ( \@ct, $tlen );
}

sub _supercompact_tree {
	my $s  = shift;
	my $ct = shift;

	my @nodes = @$ct;
	my @sct   = ();

	foreach my $node (@nodes) {

		if ( $node->{node} eq 'loop' ) {
			push @sct, $node;
		}
		elsif ( $node->{node} eq 'stem' ) {
			$node->{children} = $s->_supercompact_tree( $node->{children} );

			my @cnodes = @{ $node->{children} };
			my @loops  = grep { $_->{node} eq 'loop' } @cnodes;
			my @others = grep { $_->{node} ne 'loop' } @cnodes;

			if (    scalar(@others) == 0
				and scalar(@loops) == 1 )
			{

				my $len      = $node->{len};
				my $loop_len = $loops[0]->{len};
				my $stem_len = ( $len - $loop_len ) / 2;

				push @sct,
				  {
					node          => 'stem_loop',
					start         => $node->{start},
					stop          => $node->{stop},
					len           => $node->{len},
					loop          => $loops[0],
					loop_len      => $loop_len,
					stem_len      => $stem_len,
					left_arm_len  => $stem_len,
					right_arm_len => $stem_len,
					rloops        => [],
					lloops        => [],
				  };
			}
			elsif ( scalar(@others) == 1 ) {

				my ( $lloop, $rloop );

				if ( scalar(@loops) == 2 ) {

					$lloop = $loops[0];
					$rloop = $loops[1];

				}
				else {
					if ( $cnodes[0]->{node} eq 'loop' ) {
						$lloop = $loops[0];
					}
					else {
						$rloop = $loops[0];
					}
				}

				my $stem     = $others[0];
				my $new_node = {
					node  => $stem->{node},
					start => $node->{start},
					stop  => $node->{stop},
					len   => $node->{len}
				};

				if ( $stem->{node} eq 'stem' ) {
					$new_node->{children} = $stem->{children};

				}
				else {    # stem_loop
					$new_node->{rloops} = $stem->{rloops};

					$new_node->{lloops} = $stem->{lloops};

					push @{ $new_node->{rloops} }, $rloop
					  if $rloop;

					push @{ $new_node->{lloops} }, $lloop
					  if $lloop;

					$rloop = { len => 0 } unless $rloop;
					$lloop = { len => 0 } unless $lloop;

					$new_node->{loop} = $stem->{loop};

					$new_node->{loop_len} = $stem->{loop_len};

					my $stem_inc = (
						(
							$node->{len} -
							  $stem->{len} -
							  $rloop->{len} -
							  $lloop->{len}
						) / 2
					);

					$new_node->{stem_len} = $stem->{stem_len} + $stem_inc;

					$new_node->{left_arm_len} =
					  $stem->{left_arm_len} + $stem_inc + $lloop->{len};

					$new_node->{right_arm_len} =
					  $stem->{right_arm_len} + $stem_inc + $rloop->{len};
				}

				push @sct, $new_node;
			}
			else {
				push @sct, $node;
			}
		}
	}

	return wantarray ? @sct : \@sct;
}

sub is_good_structure {
	my $s       = shift;
	my $largest = $s->largest_loop();

	return (  ( $largest->{loop_len} >= 4 )
		  and ( $largest->{stem_len} > 10 )
		  and
		  ( $largest->{left_arm_len} > 16 or $largest->{right_arm_len} > 16 )
		  and ( $largest->{left_arm_len} < 2 * $largest->{stem_len} )
		  and ( $largest->{right_arm_len} < 2 * $largest->{stem_len} ) )
	  ? 1
	  : 0;
}

sub in_stem_loop_arm {
	my $s         = shift;
	my %a         = @_;
	my $start     = $a{start};
	my $stop      = $a{stop};
	my $tolerance = $a{tolerance} || 0;
	my $diag      = $a{diag};

	my $sloops = $s->stem_loops();
	my @sloops = @$sloops;

	@sloops = sort { $a->{start} <=> $b->{start} } @sloops;

	my $res = 0;
	foreach my $sl (@sloops) {
		next if $stop < $sl->{start};
		next if $start > $sl->{stop};

		# must start after stem-loop start
		my $startcond = ( $start >= $sl->{start} - $tolerance );

		# must stop before stem-loop stop
		my $stopcond = ( $stop <= $sl->{stop} + $tolerance );

		# is in left arm
		my $leftcond =
		  ( $stop <=
			  $sl->{start} + $sl->{left_arm_len} - 1 + ( $tolerance / 2 ) );

		# is in right arm
		my $rightcond =
		  ( $start >= $sl->{start} +
			  $sl->{left_arm_len} +
			  $sl->{loop_len} -
			  ( $tolerance / 2 ) );

		# is in either arm
		my $armcond = $leftcond || $rightcond;

		$res = 1 if $startcond && $stopcond && $armcond;

	}

	return $res;
}

sub on_largest_loop {
	my $s         = shift;
	my %a         = @_;
	my $start     = $a{start};
	my $stop      = $a{stop};
	my $tolerance = $a{tolerance} || 0;
	my $largest   = $s->largest_loop();

	return (  $start >= ( $largest->{start} - $tolerance )
		  and $stop <= ( $largest->{stop} + $tolerance ) ) ? 1 : 0;
}

sub largest_loop {
	my $s = shift;
	return $s->{largest_loop} if defined $s->{largest_loop};

	my $sloops = $s->stem_loops();
	my @sloops = @$sloops;

	my $largest;
	foreach my $sl (@sloops) {
		$largest = $sl
		  if !defined $largest
			  or $sl->{stem_len} > $largest->{stem_len};
	}

	return $s->{largest_loop} = $largest;
}

sub largest_loop_sequence {
	my $s        = shift;
	my %a        = @_;
	my $mask     = $a{mask_loop};
	my $largest  = $s->largest_loop();
	my $sequence = $s->sequence();

	if ($mask) {
		substr( $sequence, $largest->{start} + $largest->{left_arm_len},
			$largest->{loop_len} )
		  = 'S' x $largest->{loop_len};
	}

	return substr $sequence, $largest->{start}, $largest->{len};
}

sub largest_loop_structure {
	my $s         = shift;
	my $largest   = $s->largest_loop();
	my $structure = $s->structure();

	return substr $structure, $largest->{start}, $largest->{len};
}

sub stem_loops {
	my $s = shift;
	return $s->{_stem_loops} if defined $s->{_stem_loops};

	my @sloops = ();

	my $sct = $s->structure_tree();

	my @queue = @$sct;

	while (@queue) {
		my $n = shift @queue;
		next if $n->{node} eq 'loop';
		push @queue, @{ $n->{children} }
		  if $n->{node} eq 'stem';
		push @sloops, $n
		  if $n->{node} eq 'stem_loop';
	}

	$s->{_stem_loops} = \@sloops;
}

1;

