package BAOBAB::Fasta;

# $Id: Fasta.pm,v 1.8 2010/02/24 00:41:09 rnl14480 Exp $

use BAOBAB::Fasta::Sequence;
use BAOBAB::Index;
use Data::Dumper;

use strict;

sub new {
	my $c    = shift;
	my %a    = @_;
	my $file = $a{filename};
	print STDERR "Creating Fasta object for filename $file\n";
	my $r = {
		nr_sequences    => undef,
		total_size      => undef,
		_filename       => $file,
		_sequences      => $a{sequences} || [],
		_index          => undef,
		_sequence_hash  => {},
		_sequence_order => {},
		_header_handler => $a{header_handler} || sub { return shift; },
		_save_memory    => $a{save_memory}
	};

	$r = _read_file($r) if defined $file;
	bless $r, $c;

	$r->_init();

	return $r;
}

sub _read_file {
	my $r = shift;
	my $f = $r->{_filename};
	my $h = $r->{_header_handler};

	my $file;
	open( $file, $f ) or die "Cannot open file $f : $!\n";

	#print STDERR "Reading file\n";

	my $state = 0;
	my $header;
	my $sequence;
	my @sequences = ();
	while (<$file>) {
		my $line = $_;
		chomp($line);

		if ( $line =~ /^>/ ) {
			if ( defined $sequence ) {
				push @sequences, [ $header, $sequence ];
				$header = $sequence = undef;
			}
			$header = $line;
			substr( $header, 0, 1 ) = '';
		}
		elsif ( $line =~ /[atgc]+/i ) {
			$sequence .= $line;
		}
	}

	close($file);

	# wrap-up
	if ( defined $sequence ) {
		push @sequences, [ $header, $sequence ];
		$header = $sequence = undef;
	}

	my @seqobjs = ();
	my $spos    = 0;
	foreach my $s (@sequences) {
		my $seq = BAOBAB::Fasta::Sequence->new(
			header      => $s->[0],
			id          => $h->( $s->[0] ),
			sequence    => $s->[1],
			save_memory => $r->{_save_memory}
		);

		#print STDERR "header = " . $s->[0] . "\n";
		#print STDERR "id = " . $h->( $s->[0] ) . "\n";
		$r->{_sequence_hash}{ $h->( $s->[0] ) }  = $seq;
		$r->{_sequence_order}{ $h->( $s->[0] ) } = $spos++;
		push @seqobjs, $seq;
	}

	$r->{_sequences} = \@seqobjs;

	return $r;

}

sub _init {
	my $s = shift;

	return $s unless $s->sequences();

	my @seqobjs = $s->sequences();
	$s->{nr_sequences} = scalar(@seqobjs);
	$s->{total_size}   = 0;
	foreach my $o (@seqobjs) {
		$s->{total_size} += $o->length();
	}

	#print STDERR "The object is \n" . join( "\n", keys %$s ) . "\n";
	print STDERR "Creating Index object for file " . $s->filename . "\n";
	my $idx = BAOBAB::Index->new( file => $s->filename );
	$s->{_index} = $idx;

	return $s;
}

sub filename     { return (shift)->{_filename}; }
sub index        { return (shift)->{_index}; }
sub nr_sequences { return (shift)->{nr_sequences}; }
sub total_size   { return (shift)->{total_size}; }

sub sequences {
	my $s = shift;
	return wantarray ? @{ $s->{_sequences} } : $s->{_sequences};
}

sub get_sequence_from_id {
	my $s  = shift;
	my $id = shift;

	return $s->{_sequence_hash}{$id};
}

sub new_from_filter {
	my $s  = shift;
	my $fc = shift;

	my @seqs = ();
	foreach my $o ( @{ $s->{_sequences} } ) {
		push @seqs, $o
		  if $fc->($o);
	}

	return BAOBAB::Fasta->new( sequences => \@seqs );
}

sub get_region {
	my $s = shift;
	my %a = @_;

	# print STDERR "Asking for a slice of chromosome ".$a{chr}."\n";
	my $seqn = $s->{_sequence_order}{ $a{chr} };
	$a{sequence} = $seqn;

	return $s->index->get_region(%a);
}

sub print {
	my $s     = shift;
	my %a     = @_;
	my $fname = $a{filename};

	my $out = \*STDOUT;
	if ($fname) {
		open( $out, ">$fname" ) or die "Cannot open $fname : $!\n";
	}

	foreach my $o ( @{ $s->{_sequences} } ) {
		print ">" . $o->header() . "\n";
		print $o->sequence() . "\n";
	}

	close($out) if $fname;
	return;
}

1;

